/*
	RT Bench - Benchmarks various hash functions for Rainbow Tables
	Copyright (C) 2012 Steve Thomas <steve AT tobtu DOT com>

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#include "hashfactory.h"
#include "md4.h"
#include "md4help.h"

uint32_t MD4::inited = MD4::initCalc();

#ifdef ARC_x86

void MD4::calcSSE2(vector4 *pw, vector4 *hash)
{
	md4Func((__m128i*) pw, (__m128i*) hash);
}

#endif

void MD4::calcNorm(vector4 *pw, vector4 *hash)
{
	uint32_t a = 0x67452301, b = 0xefcdab89, c = 0x98badcfe, d = 0x10325476;

	#define MD4_FF(a, b, c, d, x, s) a += (((c ^ d) & b) ^ d)                    + FROM_LITTLE_ENDIAN_32(x); a = (a << s) | (a >> (32 - s));
	#define MD4_GG(a, b, c, d, x, s) a += (((b & c) | d) & (b | c)) + 0x5a827999 + FROM_LITTLE_ENDIAN_32(x); a = (a << s) | (a >> (32 - s));
	#define MD4_HH(a, b, c, d, x, s) a += (b ^ c ^ d)               + 0x6ed9eba1 + FROM_LITTLE_ENDIAN_32(x); a = (a << s) | (a >> (32 - s));

	#define x0   pw[1].data[0]
	#define x1   pw[2].data[0]
	#define x2   pw[3].data[0]
	#define x3   pw[4].data[0]
	#define x4   pw[5].data[0]
	#define x5   pw[6].data[0]
	#define x6   pw[7].data[0]
	#define x7   pw[8].data[0]
	#define x8   0
	#define x9   0
	#define x10  0
	#define x11  0
	#define x12  0
	#define x13  0
	#define x14  pw[0].data[0]
	#define x15  0

	pw[0].data[0] <<= 3;

	MD4_FF(a, b, c, d,  x0,  3); MD4_FF(d, a, b, c,  x1,  7); MD4_FF(c, d, a, b,  x2, 11); MD4_FF(b, c, d, a,  x3, 19);
	MD4_FF(a, b, c, d,  x4,  3); MD4_FF(d, a, b, c,  x5,  7); MD4_FF(c, d, a, b,  x6, 11); MD4_FF(b, c, d, a,  x7, 19);
	MD4_FF(a, b, c, d,  x8,  3); MD4_FF(d, a, b, c,  x9,  7); MD4_FF(c, d, a, b, x10, 11); MD4_FF(b, c, d, a, x11, 19);
	MD4_FF(a, b, c, d, x12,  3); MD4_FF(d, a, b, c, x13,  7); MD4_FF(c, d, a, b, x14, 11); MD4_FF(b, c, d, a, x15, 19);

	MD4_GG(a, b, c, d,  x0,  3); MD4_GG(d, a, b, c,  x4,  5); MD4_GG(c, d, a, b,  x8,  9); MD4_GG(b, c, d, a, x12, 13);
	MD4_GG(a, b, c, d,  x1,  3); MD4_GG(d, a, b, c,  x5,  5); MD4_GG(c, d, a, b,  x9,  9); MD4_GG(b, c, d, a, x13, 13);
	MD4_GG(a, b, c, d,  x2,  3); MD4_GG(d, a, b, c,  x6,  5); MD4_GG(c, d, a, b, x10,  9); MD4_GG(b, c, d, a, x14, 13);
	MD4_GG(a, b, c, d,  x3,  3); MD4_GG(d, a, b, c,  x7,  5); MD4_GG(c, d, a, b, x11,  9); MD4_GG(b, c, d, a, x15, 13);

	MD4_HH(a, b, c, d,  x0,  3); MD4_HH(d, a, b, c,  x8,  9); MD4_HH(c, d, a, b,  x4, 11); MD4_HH(b, c, d, a, x12, 15);
	MD4_HH(a, b, c, d,  x2,  3); MD4_HH(d, a, b, c, x10,  9); MD4_HH(c, d, a, b,  x6, 11); MD4_HH(b, c, d, a, x14, 15);
	MD4_HH(a, b, c, d,  x1,  3); MD4_HH(d, a, b, c,  x9,  9); MD4_HH(c, d, a, b,  x5, 11); MD4_HH(b, c, d, a, x13, 15);
	MD4_HH(a, b, c, d,  x3,  3); MD4_HH(d, a, b, c, x11,  9); MD4_HH(c, d, a, b,  x7, 11); MD4_HH(b, c, d, a, x15, 15);

	#undef MD4_FF
	#undef MD4_GG
	#undef MD4_HH

	#undef x0
	#undef x1
	#undef x2
	#undef x3
	#undef x4
	#undef x5
	#undef x6
	#undef x7
	#undef x8
	#undef x9
	#undef x10
	#undef x11
	#undef x12
	#undef x13
	#undef x14
	#undef x15

	hash[0].data[0] = TO_LITTLE_ENDIAN_32(a + 0x67452301);
	hash[1].data[0] = TO_LITTLE_ENDIAN_32(b + 0xefcdab89);
	hash[2].data[0] = TO_LITTLE_ENDIAN_32(c + 0x98badcfe);
	hash[3].data[0] = TO_LITTLE_ENDIAN_32(d + 0x10325476);
}

uint32_t MD4::initCalc()
{
#ifdef ARC_x86
	if ((getInstructionSets() & (IS_AVX | IS_XOP)) == (IS_AVX | IS_XOP))
	{
		return HashFactory::registerHash("md4", QUOTE(__MD4_INTERLACED__) "xAVX_XOP", 0, 31, MD4::calcAVXXOP,  16, 4 * __MD4_INTERLACED__, CE_ASCII_BIN8 | CE_80_NULL_PADDED, 8);
	}
	else if ((getInstructionSets() & IS_AVX) == IS_AVX)
	{
		return HashFactory::registerHash("md4", QUOTE(__MD4_INTERLACED__) "xAVX",     0, 31, MD4::calcAVX,     16, 4 * __MD4_INTERLACED__, CE_ASCII_BIN8 | CE_80_NULL_PADDED, 8);
	}
	else if ((getInstructionSets() & IS_SSE2) == IS_SSE2)
	{
		return HashFactory::registerHash("md4", QUOTE(__MD4_INTERLACED__) "xSSE2",    0, 31, MD4::calcSSE2,    16, 4 * __MD4_INTERLACED__, CE_ASCII_BIN8 | CE_80_NULL_PADDED, 8);
	}
#endif
	return HashFactory::registerHash("md4", "scalar",                             0, 31, MD4::calcNorm,    16,                      1, CE_ASCII_BIN8 | CE_80_NULL_PADDED, 8);
}
