/*
	RT Bench - Benchmarks various hash functions for Rainbow Tables
	Copyright (C) 2012 Steve Thomas <steve AT tobtu DOT com>

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#include "hashfactory.h"
#include "sha1.h"
#include "sha1help.h"

uint32_t SHA1::inited = SHA1::initCalc();

#ifdef ARC_x86

void SHA1::calcSSSE3(vector4 *pw, vector4 *hash)
{
	sha1Func_SSSE3((__m128i*) pw, (__m128i*) hash);
}

void SHA1::calcSSE2(vector4 *pw, vector4 *hash)
{
	sha1Func((__m128i*) pw, (__m128i*) hash);
}

#endif

void SHA1::calcNorm(vector4 *pw, vector4 *hash)
{
	uint32_t a = 0x67452301, b = 0xefcdab89, c = 0x98badcfe, d = 0x10325476, e = 0xc3d2e1f0;

	#define ROT(n,s) ((n << s) | (n >> (32 - s)))

	#define SHA1_FF(a,b,c,d,e,x,i)      e += (((c ^ d) & b) ^ d)       + x(i) + 0x5a827999 + ROT(a, 5); b = ROT(b, 30);
	#define SHA1_GG(a,b,c,d,e,x,i)      e += (b ^ (c ^ d))             + x(i) + 0x6ed9eba1 + ROT(a, 5); b = ROT(b, 30);
	#define SHA1_HH(a,b,c,d,e,x,i)      e += (((b | c) & d) | (b & c)) + x(i) + 0x8f1bbcdc + ROT(a, 5); b = ROT(b, 30);
	#define SHA1_II(a,b,c,d,e,x,i)      e += (b ^ (c ^ d))             + x(i) + 0xca62c1d6 + ROT(a, 5); b = ROT(b, 30);

	#define x0(i)    (y(0) = FROM_BIG_ENDIAN_32(pw[1].data[0]))
	#define x1(i)    (y(1) = FROM_BIG_ENDIAN_32(pw[2].data[0]))
	#define x2(i)    (y(2) = FROM_BIG_ENDIAN_32(pw[3].data[0]))
	#define x3(i)    (y(3) = FROM_BIG_ENDIAN_32(pw[4].data[0]))
	#define x4(i)    (y(4) = FROM_BIG_ENDIAN_32(pw[5].data[0]))
	#define x5(i)    (y(5) = FROM_BIG_ENDIAN_32(pw[6].data[0]))
	#define x6(i)    0
	#define x7(i)    0
	#define x8(i)    0
	#define x9(i)    0
	#define x10(i)   0
	#define x11(i)   0
	#define x12(i)   0
	#define x13(i)   0
	#define x14(i)   0
	#define x15(i)   y(15)

	#define y(i)     ((uint32_t*)hash)[(i) & 15]

	#define x(i)     (y( i) = ROT((((y( i) ^ y(i+2)) ^ y(i+8)) ^ y(i+13)), 1))
	#define x0_2(i)  (y( 0) = ROT((((y( 0) ^  y( 2)) ^      0) ^       0), 1))
	#define x1_2(i)  (y( 1) = ROT((((y( 1) ^  y( 3)) ^      0) ^       0), 1))
	#define x2_2(i)  (y( 2) = ROT((((y( 2) ^  y( 4)) ^      0) ^   y(15)), 1))
	#define x3_2(i)  (y( 3) = ROT((((y( 3) ^  y( 5)) ^      0) ^   y( 0)), 1))
	#define x4_2(i)  (y( 4) = ROT((((y( 4) ^      0) ^      0) ^   y( 1)), 1))
	#define x5_2(i)  (y( 5) = ROT((((y( 5) ^      0) ^      0) ^   y( 2)), 1))
	#define x6_2(i)  (y( 6) = ROT((((    0 ^      0) ^      0) ^   y( 3)), 1))
	#define x7_2(i)  (y( 7) = ROT((((    0 ^      0) ^  y(15)) ^   y( 4)), 1))
	#define x8_2(i)  (y( 8) = ROT((((    0 ^      0) ^  y( 0)) ^   y( 5)), 1))
	#define x9_2(i)  (y( 9) = ROT((((    0 ^      0) ^  y( 1)) ^   y( 6)), 1))
	#define x10_2(i) (y(10) = ROT((((    0 ^      0) ^  y( 2)) ^   y( 7)), 1))
	#define x11_2(i) (y(11) = ROT((((    0 ^      0) ^  y( 3)) ^   y( 8)), 1))
	#define x12_2(i) (y(12) = ROT((((    0 ^      0) ^  y( 4)) ^   y( 9)), 1))
	#define x13_2(i) (y(13) = ROT((((    0 ^  y(15)) ^  y( 5)) ^   y(10)), 1))
	#define x14_2(i) (y(14) = ROT((((    0 ^  y( 0)) ^  y( 6)) ^   y(11)), 1))
	#define x15_2(i) (y(15) = ROT((((y(15) ^  y( 1)) ^  y( 7)) ^   y(12)), 1))

	y(15) = pw[0].data[0] << 3;

	SHA1_FF(a,b,c,d,e,   x0, 0); SHA1_FF(e,a,b,c,d,   x1, 1); SHA1_FF(d,e,a,b,c,   x2, 2); SHA1_FF(c,d,e,a,b,   x3, 3); SHA1_FF(b,c,d,e,a,   x4, 4);
	SHA1_FF(a,b,c,d,e,   x5, 5); SHA1_FF(e,a,b,c,d,   x6, 6); SHA1_FF(d,e,a,b,c,   x7, 7); SHA1_FF(c,d,e,a,b,   x8, 8); SHA1_FF(b,c,d,e,a,   x9, 9);
	SHA1_FF(a,b,c,d,e,  x10,10); SHA1_FF(e,a,b,c,d,  x11,11); SHA1_FF(d,e,a,b,c,  x12,12); SHA1_FF(c,d,e,a,b,  x13,13); SHA1_FF(b,c,d,e,a,  x14,14);
	SHA1_FF(a,b,c,d,e,  x15,15); SHA1_FF(e,a,b,c,d, x0_2,16); SHA1_FF(d,e,a,b,c, x1_2,17); SHA1_FF(c,d,e,a,b, x2_2,18); SHA1_FF(b,c,d,e,a, x3_2,19);

	SHA1_GG(a,b,c,d,e, x4_2,20); SHA1_GG(e,a,b,c,d, x5_2,21); SHA1_GG(d,e,a,b,c, x6_2,22); SHA1_GG(c,d,e,a,b, x7_2,23); SHA1_GG(b,c,d,e,a, x8_2,24);
	SHA1_GG(a,b,c,d,e, x9_2,25); SHA1_GG(e,a,b,c,d,x10_2,26); SHA1_GG(d,e,a,b,c,x11_2,27); SHA1_GG(c,d,e,a,b,x12_2,28); SHA1_GG(b,c,d,e,a,x13_2,29);
	SHA1_GG(a,b,c,d,e,x14_2,30); SHA1_GG(e,a,b,c,d,x15_2,31); SHA1_GG(d,e,a,b,c,    x,32); SHA1_GG(c,d,e,a,b,    x,33); SHA1_GG(b,c,d,e,a,    x,34);
	SHA1_GG(a,b,c,d,e,    x,35); SHA1_GG(e,a,b,c,d,    x,36); SHA1_GG(d,e,a,b,c,    x,37); SHA1_GG(c,d,e,a,b,    x,38); SHA1_GG(b,c,d,e,a,    x,39);

	SHA1_HH(a,b,c,d,e,    x,40); SHA1_HH(e,a,b,c,d,    x,41); SHA1_HH(d,e,a,b,c,    x,42); SHA1_HH(c,d,e,a,b,    x,43); SHA1_HH(b,c,d,e,a,    x,44);
	SHA1_HH(a,b,c,d,e,    x,45); SHA1_HH(e,a,b,c,d,    x,46); SHA1_HH(d,e,a,b,c,    x,47); SHA1_HH(c,d,e,a,b,    x,48); SHA1_HH(b,c,d,e,a,    x,49);
	SHA1_HH(a,b,c,d,e,    x,50); SHA1_HH(e,a,b,c,d,    x,51); SHA1_HH(d,e,a,b,c,    x,52); SHA1_HH(c,d,e,a,b,    x,53); SHA1_HH(b,c,d,e,a,    x,54);
	SHA1_HH(a,b,c,d,e,    x,55); SHA1_HH(e,a,b,c,d,    x,56); SHA1_HH(d,e,a,b,c,    x,57); SHA1_HH(c,d,e,a,b,    x,58); SHA1_HH(b,c,d,e,a,    x,59);

	SHA1_II(a,b,c,d,e,    x,60); SHA1_II(e,a,b,c,d,    x,61); SHA1_II(d,e,a,b,c,    x,62); SHA1_II(c,d,e,a,b,    x,63); SHA1_II(b,c,d,e,a,    x,64);
	SHA1_II(a,b,c,d,e,    x,65); SHA1_II(e,a,b,c,d,    x,66); SHA1_II(d,e,a,b,c,    x,67); SHA1_II(c,d,e,a,b,    x,68); SHA1_II(b,c,d,e,a,    x,69);
	SHA1_II(a,b,c,d,e,    x,70); SHA1_II(e,a,b,c,d,    x,71); SHA1_II(d,e,a,b,c,    x,72); SHA1_II(c,d,e,a,b,    x,73); SHA1_II(b,c,d,e,a,    x,74);
	SHA1_II(a,b,c,d,e,    x,75); SHA1_II(e,a,b,c,d,    x,76); SHA1_II(d,e,a,b,c,    x,77); SHA1_II(c,d,e,a,b,    x,78); SHA1_II(b,c,d,e,a,    x,79);

	hash[0].data[0] = TO_BIG_ENDIAN_32(a + 0x67452301);
	hash[1].data[0] = TO_BIG_ENDIAN_32(b + 0xefcdab89);
	hash[2].data[0] = TO_BIG_ENDIAN_32(c + 0x98badcfe);
	hash[3].data[0] = TO_BIG_ENDIAN_32(d + 0x10325476);
	hash[4].data[0] = TO_BIG_ENDIAN_32(d + 0xc3d2e1f0);

	#undef ROT

	#undef x0
	#undef x1
	#undef x2
	#undef x3
	#undef x4
	#undef x5
	#undef x6
	#undef x7
	#undef x8
	#undef x9
	#undef x10
	#undef x11
	#undef x12
	#undef x13
	#undef x14
	#undef x15

	#undef x
	#undef y

	#undef x0_2
	#undef x1_2
	#undef x2_2
	#undef x3_2
	#undef x4_2
	#undef x5_2
	#undef x6_2
	#undef x7_2
	#undef x8_2
	#undef x9_2
	#undef x10_2
	#undef x11_2
	#undef x12_2
	#undef x13_2
	#undef x14_2
	#undef x15_2

	#undef SHA1_FF
	#undef SHA1_GG
	#undef SHA1_HH
	#undef SHA1_II
}

uint32_t SHA1::initCalc()
{
#ifdef ARC_x86
	if ((getInstructionSets() & (IS_AVX | IS_XOP)) == (IS_AVX | IS_XOP))
	{
		return HashFactory::registerHash("sha1", QUOTE(__SHA1_INTERLACED__) "xAVX_XOP", 0, 23, SHA1::calcAVXXOP,  20, 4 * __SHA1_INTERLACED__, CE_ASCII_BIN8 | CE_80_NULL_PADDED, 6, 11);
	}
	else if ((getInstructionSets() & IS_AVX) == IS_AVX)
	{
		return HashFactory::registerHash("sha1", QUOTE(__SHA1_INTERLACED__) "xAVX",     0, 23, SHA1::calcAVX,     20, 4 * __SHA1_INTERLACED__, CE_ASCII_BIN8 | CE_80_NULL_PADDED, 6, 11);
	}
	else if ((getInstructionSets() & IS_SSSE3) == IS_SSSE3)
	{
		return HashFactory::registerHash("sha1", QUOTE(__SHA1_INTERLACED__) "xSSSE3",   0, 23, SHA1::calcSSSE3,   20, 4 * __SHA1_INTERLACED__, CE_ASCII_BIN8 | CE_80_NULL_PADDED, 6, 11);
	}
	else if ((getInstructionSets() & IS_SSE2) == IS_SSE2)
	{
		return HashFactory::registerHash("sha1", QUOTE(__SHA1_INTERLACED__) "xSSE2",    0, 23, SHA1::calcSSE2,    20, 4 * __SHA1_INTERLACED__, CE_ASCII_BIN8 | CE_80_NULL_PADDED, 6, 11);
	}
#endif
	return HashFactory::registerHash("sha1", "scalar",                              0, 23, SHA1::calcNorm,    20,                       1, CE_ASCII_BIN8 | CE_80_NULL_PADDED, 6);
}
